/*
 * Copyright (c) 2005-2007 Ondrej Palkovsky, Michal Kebrt, Petr Stepan
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/** @addtogroup arm32
 * @{
 */
/** @file
 */

#include <interrupt.h>
#include <ipc/irq.h>
#include <console/chardev.h>
#include <arch/drivers/gxemul.h>
#include <console/console.h>
#include <sysinfo/sysinfo.h>
#include <print.h>
#include <ddi/device.h>

/** Address of devices. */
#define GXEMUL_VIDEORAM		0x10000000
#define GXEMUL_KBD_ADDRESS	0x10000000
#define GXEMUL_RTC		0x15000000
#define GXEMUL_RTC_FREQ		0x15000100
#define GXEMUL_RTC_ACK		0x15000110
#define GXEMUL_IRQC		0x16000000
#define GXEMUL_IRQC_MASK	0x16000004
#define GXEMUL_IRQC_UNMASK	0x16000008

/** IRQs */
#define GXEMUL_KBD_IRQ		2
#define GXEMUL_TIMER_IRQ	4

static chardev_t console;
static irq_t gxemul_irq;
static irq_t gxemul_timer_irq;

static void gxemul_write(chardev_t *dev, const char ch);
static void gxemul_enable(chardev_t *dev);
static void gxemul_disable(chardev_t *dev);
static char gxemul_do_read(chardev_t *dev);

static chardev_operations_t gxemul_ops = {
	.resume = gxemul_enable,
	.suspend = gxemul_disable,
	.write = gxemul_write,
	.read = gxemul_do_read,
};

/** Putchar that works with gxemul */
void gxemul_write(chardev_t *dev, const char ch)
{
	*((char *) GXEMUL_VIDEORAM) = ch;
}

/* Called from getc(). */
void gxemul_enable(chardev_t *dev)
{
//	cp0_unmask_int(GXEMUL_KBD_IRQ);
	gxemul_irqc_unmask(GXEMUL_KBD_IRQ);
}

/* Called from getc(). */
void gxemul_disable(chardev_t *dev)
{
//	cp0_mask_int(GXEMUL_KBD_IRQ);
	gxemul_irqc_mask(GXEMUL_KBD_IRQ);
}

/** Read character using polling, assume interrupts disabled */
static char gxemul_do_read(chardev_t *dev)
{
	char ch;

	while (1) {
		ch = *((volatile char *) GXEMUL_KBD_ADDRESS);
		if (ch) {
			if (ch == '\r')
				return '\n';
			if (ch == 0x7f)
				return '\b';
			return ch;
		}
	}
}

/** Process keyboard interrupt. */
static void gxemul_irq_handler(irq_t *irq, void *arg, ...)
{
	if ((irq->notif_cfg.notify) && (irq->notif_cfg.answerbox))
		ipc_irq_send_notif(irq);
	else {
		char ch = 0;
		
			ch = *((char *) GXEMUL_KBD_ADDRESS);
			if (ch =='\r')
				ch = '\n';
			if (ch == 0x7f)
				ch = '\b';
			//chardev_push_character(&console, ch);
			printf("%c", ch);
	}
}

static irq_ownership_t gxemul_claim(void)
{
	return IRQ_ACCEPT;
}

void gxemul_kbd_grab(void)
{
	ipl_t ipl = interrupts_disable();
	spinlock_lock(&msim_irq.lock);
	gxemul_irq.notif_cfg.notify = false;
	spinlock_unlock(&gxemul_irq.lock);
	interrupts_restore(ipl);
}

void gxemul_kbd_release(void)
{
	ipl_t ipl = interrupts_disable();
	spinlock_lock(&gxemul_irq.lock);
	if (gxemul_irq.notif_cfg.answerbox)
		gxemul_irq.notif_cfg.notify = true;
	spinlock_unlock(&gxemul_irq.lock);
	interrupts_restore(ipl);
}


/** Return console object representing msim console */
void gxemul_console(devno_t devno)
{
	chardev_initialize("msim_console", &console, &gxemul_ops);
	stdin = &console;
	stdout = &console;
	
	irq_initialize(&gxemul_irq);
	gxemul_irq.devno = devno;
	gxemul_irq.inr = GXEMUL_KBD_IRQ;
	gxemul_irq.claim = gxemul_claim;
	gxemul_irq.handler = gxemul_irq_handler;
	irq_register(&gxemul_irq);
	
//	cp0_unmask_int(GXEMUL_KBD_IRQ);
	gxemul_irqc_unmask(GXEMUL_KBD_IRQ);
	
	sysinfo_set_item_val("kbd", NULL, true);
	sysinfo_set_item_val("kbd.devno", NULL, devno);
	sysinfo_set_item_val("kbd.inr", NULL, GXEMUL_KBD_IRQ);
	sysinfo_set_item_val("kbd.address.virtual", NULL, GXEMUL_KBD_ADDRESS);
}

/** Return the mask of active interrupts. */
inline uint32_t gxemul_irqc_get_sources(void)
{
	return *(uint32_t*) GXEMUL_IRQC;
}

/** Masks interrupt.
 * 
 * @param irq interrupt number
 */
inline void gxemul_irqc_mask(uint32_t irq)
{
	*(uint32_t*) GXEMUL_IRQC_MASK = irq;
}

/** Unmasks interrupt.
 * 
 * @param irq interrupt number
 */
inline void gxemul_irqc_unmask(uint32_t irq)
{
	*(uint32_t*) GXEMUL_IRQC_UNMASK = irq;
}


/** Starts gxemul Real Time Clock device, which asserts regular interrupts.
 * 
 * @param frequency interrupts frequency (0 disables RTC)
 */
void gxemul_timer_start(uint32_t frequency)
{
	*(uint32_t*) GXEMUL_RTC_FREQ = frequency;
}

static irq_ownership_t gxemul_timer_claim(void)
{
	return IRQ_ACCEPT;
}

static void gxemul_timer_irq_handler(irq_t *irq, void *arg, ...)
{
	/* TODO time drifts ??
	unsigned long drift;

	drift = cp0_count_read() - nextcount;
	while (drift > cp0_compare_value) {
		drift -= cp0_compare_value;
		CPU->missed_clock_ticks++;
	}
	nextcount = cp0_count_read() + cp0_compare_value - drift;
	cp0_compare_write(nextcount);
	*/

	/*
	* We are holding a lock which prevents preemption.
	* Release the lock, call clock() and reacquire the lock again.
	*/
	spinlock_unlock(&irq->lock);
	//clock();
	puts(" ");
	spinlock_lock(&irq->lock);
	
	/* TODO what's that? *
	if (virtual_timer_fnc != NULL)
		virtual_timer_fnc();
	*/
}

/**
 * Initializes and registers timer interrupt handler.
 */
void gxemul_timer_irq_init()
{
	irq_initialize(&gxemul_timer_irq);
	gxemul_timer_irq.devno = device_assign_devno();
	gxemul_timer_irq.inr = GXEMUL_TIMER_IRQ;
	gxemul_timer_irq.claim = gxemul_timer_claim;
	gxemul_timer_irq.handler = gxemul_timer_irq_handler;
	irq_register(&gxemul_timer_irq);
}


/** @}
 */
